/* VzImagination (C) 2018-2019 Stephane Charette <stephanecharette@gmail.com>
 * $Id: TestMatManipResize.cpp 2746 2019-04-10 02:58:03Z stephane $
 */

#include "TestUtil.hpp"


TEST(Resize, Invalid)
{
	cv::Mat src;

	// both parms cannot be zero
	ASSERT_THROW(vz::resize(src, 0), std::exception);
	ASSERT_THROW(vz::resize(src, 0, 0), std::exception);

	// nor can the parms be < 0
	ASSERT_THROW(vz::resize(src, -1), std::exception);

	// nor can the image be empty
	ASSERT_THROW(vz::resize(src, 640, 480), std::exception);
}


TEST(Resize, SmallerWidth)
{
	for (const auto & [path, name] : complex_test_images())
	{
		cv::Mat src = cv::imread(path);
		ASSERT_NE(src.cols, 0);
		ASSERT_NE(src.rows, 0);

		// CAREFUL!  Even though we think we're dividing by 4.0, the resulting width is then rounded to the nearest
		// integer, meaning the ratio we're using may be slightly different than what we think it was meant to be.
		const int new_width = std::round((double)src.cols / 4.0);
		const double actual_ratio = (double)new_width / (double)src.cols;
		const int expected_height = std::round((double)src.rows * actual_ratio);

		std::cout << "resizing " << name << " (original size is " << src.cols << "x" << src.rows << ", new width will be " << new_width << ", ratio is " << actual_ratio << ", hopefully the height will be " << expected_height << ")" << std::endl;

		cv::Mat dst1 = vz::resize(src, new_width);
		ASSERT_EQ(dst1.cols, new_width);
		ASSERT_EQ(dst1.rows, expected_height);

		cv::Mat dst2 = vz::resize(src, actual_ratio);
		ASSERT_EQ(dst1.cols, dst2.cols);
		ASSERT_EQ(dst1.rows, dst2.rows);

		// regardless of what method was used to generate the images, they should be exactly the same
		ASSERT_TRUE(vz::compare_exact(dst1, dst2));

		cv::imwrite(name + "_resized_" + std::to_string(dst1.cols) + "x" + std::to_string(dst1.rows) + ".jpg", dst1, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
	}
}


TEST(Resize, SmallerHeight)
{
	// similar to the "SmallerWidth" test above, but change the height instead of the width

	for (const auto & [path, name] : complex_test_images())
	{
		cv::Mat src = cv::imread(path);
		ASSERT_NE(src.cols, 0);
		ASSERT_NE(src.rows, 0);

		// the test above resizes images 25%, with this one we'll use 30% to get slightly different numbers
		const int new_height = std::round((double)src.rows * 0.3);
		const double actual_ratio = (double)new_height / (double)src.rows;
		const int expected_width = std::round((double)src.cols * actual_ratio);

		std::cout << "resizing " << name << " (original size is " << src.cols << "x" << src.rows << ", new height will be " << new_height << ", ratio is " << actual_ratio << ", hopefully the width will be " << expected_width << ")" << std::endl;

		cv::Mat dst1 = vz::resize(src, 0, new_height);
		ASSERT_EQ(dst1.cols, expected_width);
		ASSERT_EQ(dst1.rows, new_height);

		cv::Mat dst2 = vz::resize(src, actual_ratio);
		ASSERT_EQ(dst1.cols, dst2.cols);
		ASSERT_EQ(dst1.rows, dst2.rows);

		// regardless of what method was used to generate the images, they should be exactly the same
		ASSERT_TRUE(vz::compare_exact(dst1, dst2));

		cv::imwrite(name + "_resized_" + std::to_string(dst1.cols) + "x" + std::to_string(dst1.rows) + ".jpg", dst1, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
	}
}


TEST(Resize, InvalidSize)
{
	for (const auto & [path, name] : complex_test_images())
	{
		cv::Mat src = cv::imread(path);
		ASSERT_NE(src.cols, 0);
		ASSERT_NE(src.rows, 0);

		cv::Mat dst = vz::resize(src, 0.0001); // largest image is 3384 pixels wide, so 3384 * 0.0001 = 0.3
		ASSERT_TRUE(dst.empty());
		ASSERT_EQ(dst.cols, 0);
		ASSERT_EQ(dst.rows, 0);

		dst = vz::resize(src, 0.0);
		ASSERT_TRUE(dst.empty());
		ASSERT_EQ(dst.cols, 0);
		ASSERT_EQ(dst.rows, 0);
	}
}


TEST(Resize, LargerWidth)
{
	for (const auto & [path, name] : complex_test_images())
	{
		cv::Mat src = cv::imread(path);
		ASSERT_NE(src.cols, 0);
		ASSERT_NE(src.rows, 0);

		double scale_factor = 1.33333;
		const int new_width = std::round( double(src.cols) * scale_factor);

		// because we round the width to an integer, the scale factor we use isn't necessarily
		// what we initially expected it to be, so calculate the real scale
		scale_factor = double(new_width) / double(src.cols);

		const int expected_height = std::round( double(src.rows) * scale_factor);

		std::cout << "resizing " << name << " (original size is " << src.cols << "x" << src.rows << ", scale is " << scale_factor << ", new size will be " << new_width << "x" << expected_height << ")" << std::endl;

		cv::Mat dst1 = vz::resize(src, new_width);
		ASSERT_EQ(dst1.cols, new_width);
		ASSERT_EQ(dst1.rows, expected_height);

		cv::Mat dst2 = vz::resize(src, scale_factor);
		ASSERT_EQ(dst1.cols, dst2.cols);
		ASSERT_EQ(dst1.rows, dst2.rows);

		// regardless of what method was used to generate the images, they should be exactly the same
		ASSERT_TRUE(vz::compare_exact(dst1, dst2));

		cv::imwrite(name + "_resized_" + std::to_string(dst1.cols) + "x" + std::to_string(dst1.rows) + ".jpg", dst1, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
	}
}


TEST(Resize, LargerHeight)
{
	for (const auto & [path, name] : complex_test_images())
	{
		cv::Mat src = cv::imread(path);
		ASSERT_NE(src.cols, 0);
		ASSERT_NE(src.rows, 0);

		double scale_factor = 1.7654321;
		const int new_height = std::round( double(src.rows) * scale_factor);

		// because we round the width to an integer, the scale factor we use isn't necessarily
		// what we initially expected it to be, so calculate the real scale
		scale_factor = double(new_height) / double(src.rows);

		const int expected_width = std::round( double(src.cols) * scale_factor);

		std::cout << "resizing " << name << " (original size is " << src.cols << "x" << src.rows << ", scale is " << scale_factor << ", new size will be " << expected_width << "x" << new_height << ")" << std::endl;

		cv::Mat dst1 = vz::resize(src, 0, new_height);
		ASSERT_EQ(dst1.cols, expected_width);
		ASSERT_EQ(dst1.rows, new_height);

		cv::Mat dst2 = vz::resize(src, scale_factor);
		ASSERT_EQ(dst1.cols, dst2.cols);
		ASSERT_EQ(dst1.rows, dst2.rows);

		// regardless of what method was used to generate the images, they should be exactly the same
		ASSERT_TRUE(vz::compare_exact(dst1, dst2));

		cv::imwrite(name + "_resized_" + std::to_string(dst1.cols) + "x" + std::to_string(dst1.rows) + ".jpg", dst1, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
	}
}


TEST(Resize, Letterbox)
{
	for (const auto & [path, name] : complex_test_images())
	{
		cv::Mat src = cv::imread(path);

		cv::Mat dst = vz::letterbox(src, 640, 400);
		ASSERT_EQ(dst.cols, 640);
		ASSERT_EQ(dst.rows, 400);

		cv::imwrite(name + "_letterbox_640x400.jpg", dst, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});

		dst = vz::letterbox(src, 999, 2002);
		ASSERT_EQ(dst.cols, 999);
		ASSERT_EQ(dst.rows, 2002);

		cv::imwrite(name + "_letterbox_999x2002.jpg", dst, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
	}
}
